#!/usr/bin/perl
use strict;
use warnings;
use utf8;

use File::Spec;
use Scalar::Util 'looks_like_number';
use Getopt::Long;
use FindBin '$Bin';

use Gtk2 '-init';
use Gnome2::Canvas;

use lib $Bin;
use colortable;
use history;

use Data::Dumper;

my $max_iter;
my $max_iter_max;

sub BEGIN
{
  my $use_percent=0;
  if(eval{ require Inline; })
  {
    $max_iter=5000;
    $max_iter_max=100_000;
    Inline->import(C => 'DATA');
    $colortable::use_percent=$use_percent;
    if($use_percent)
    { *iterate_mandelbrot=\&_iterate_mandelbrot_double_c; }
    else
    { *iterate_mandelbrot=\&_iterate_mandelbrot_long_c; }
  }
  else
  {
    $max_iter=300;
    $max_iter_max=10000;
    *iterate_mandelbrot=\&_iterate_mandelbrot_perl;
    $colortable::use_percent=$use_percent;
  }
}

my $width=600;
my $height=600;
my $zoom=200;
my $shift_x=2;
my $shift_y=1.5;
my $forks=1;

my $result = GetOptions (
  "width=i"  => \$width,
  "height=i" => \$height,
  "xpos=f"   => \$shift_x,
  "ypos=f"   => \$shift_y,
  "zoom=f"   => \$zoom,
  "iter=i"   => \$max_iter,
  "forks=i"  => \$forks,
  "help"     => sub{ help(); }
);


help("UNKNOWN Option: @ARGV") unless($result && !@ARGV);

sub help
{
  my $msg=shift;
  my $out=\*STDOUT;
  if($msg)
  {
    $out=\*STDERR;
    print $out "ERROR: $msg\n";
  }
  print $out <<EOH;
Usage: $0 [--width=<INT>] [--height=<INT>] [--xpos=<FLOAT>] [--ypos=<FLOAT>] [--zoom=<FLOAT>] [--iter=<INT>] [--forks=<INT>]
  width  - width of image
  height - height of image
  xpos   - X Position in Fractal
  ypos   - Y Position in Fractal
  zoom   - zoom in
  iter   - precision of calculation
  forks  - render in N processes
EOH
  exit(1) if($msg);
  exit(0);
}

my $bit_per_color=8;
my $pack="C3";

my $raw_pack='L';
$raw_pack='F' if($colortable::use_percent);

colortable->new($bit_per_color,$pack,$raw_pack)->set_default();
history->new([qw(pixmap raw shift_x shift_y zoom max_iter)])->set_default();
{
  my $val='';
  $val=pack("$raw_pack*",((0,0,0)x$width*$height));
  #for(0..$width*$height){ $val.=pack($raw_pack,0); }
  my $pixbuf=Gtk2::Gdk::Pixbuf->new('GDK_COLORSPACE_RGB', 0, $bit_per_color, $width, $height);
  history->add($pixbuf, $val, $shift_x, $shift_y, $zoom, $max_iter);
}
my $file;
my $change=0;

my $window = Gtk2::Window->new;
$window->signal_connect (delete_event => \&do_exit);
$window->set_title('Mandelbrot');

my $vbox = Gtk2::VBox->new(0,0);
$window->add($vbox);

my $hbox = Gtk2::HBox->new(0,0);
$vbox->pack_start($hbox, 0, 1, 5);
$hbox->pack_start(Gtk2::Label->new(' HISTORY '), 0, 1, 5);

my $bnz=Gtk2::Button->new_with_label('ZURÜCK');
$hbox->pack_start($bnz, 1, 1, 5);
$bnz->set_sensitive(0);

my $bnv=Gtk2::Button->new_with_label('VOR');
$hbox->pack_start($bnv, 1, 1, 5);
$bnv->set_sensitive(0);

my $txt=Gtk2::Label->new('');
$vbox->pack_start($txt, 0, 1, 0);

my $swin=Gtk2::ScrolledWindow->new();
$vbox->pack_start($swin, 1,1,0);
$swin->set_placement('GTK_CORNER_TOP_LEFT');
$swin->set_size_request(520,520);

my $canvas = Gnome2::Canvas->new_aa();
$canvas->set_size_request($width,$height);
$canvas->set_scroll_region (0, 0, $width, $height);
$swin->add($canvas);

my $root = $canvas->root;
my $sitem = Gnome2::Canvas::Item->new($root,'Gnome2::Canvas::Rect',
                            x1 => 0, y1 => 0, y2 => $height, x2 => $width,
                            fill_color_rgba => 0xFFFFFF30,
                            outline_color => 'black',
                          );
my $pitem = Gnome2::Canvas::Item->new($root, 'Gnome2::Canvas::Pixbuf',
                            x => 0, y => 0, height=>$height, width=>$width,
                            pixbuf=>history->get(undef,0),
                          );
$sitem->signal_connect("event", sub{ do_select($pitem,$_[0],$_[1]) });
$pitem->signal_connect("event", sub{ do_select($_[0],$sitem,$_[1]) });
$pitem->{HAS_CHANGED}=1;

my $hbox2 = Gtk2::HBox->new(0,0);
$vbox->pack_start($hbox2, 0, 1, 5);
$hbox2->pack_start(Gtk2::Label->new(' Genauigkeit '), 0, 1, 5);

my $deep=Gtk2::Adjustment->new ($max_iter, 0, $max_iter_max, 100, 1 , 1);
my $sb=Gtk2::HScrollbar->new($deep);
$sb->set_show_fill_level(1);
$hbox2->pack_start($sb, 1, 1, 0);

my $hbox4 = Gtk2::HBox->new(0,0);
$vbox->pack_start($hbox4, 0, 1, 5);

my $b1=Gtk2::Button->new_with_label('BERECHNEN');
$hbox4->pack_start($b1, 1, 1, 5);
$b1->signal_connect('clicked' => sub{
                                      do_render(shift,$pitem,$sitem,$bnv,$bnz,$txt,$deep,
                                                $width,$height,$shift_x,$shift_y,
                                                $zoom,$max_iter,$bit_per_color,$forks,$raw_pack);
                                      $max_iter=int($deep->value());
                                    });

my $bk=Gtk2::Button->new_with_label('Kooradinate');
$hbox4->pack_start($bk, 0, 1, 5);
$bk->signal_connect('clicked' => sub{
    my $vv=[(history->get())[2..5]];

    my $sx=($width/$vv->[2])/2;
    my $sy=($height/$vv->[2])/2;
    $vv->[0]-=$sx;
    $vv->[1]-=$sy;

    my ($vals,$mod)=do_edit_koordinates($window,$vv);
    if($mod)
    {
      #print "@$vv => @$vals\n";

      my $sx=($width/$vals->[2])/2;
      my $sy=($height/$vals->[2])/2;
      $vals->[0]+=$sx;
      $vals->[1]+=$sy;

      ($shift_x,$shift_y,$zoom,$max_iter)=@$vals;
      $deep->value($max_iter);
      $pitem->{HAS_CHANGED}=1;
    }
  });

my $bnf=Gtk2::Button->new_with_label('Farben');
$hbox4->pack_start($bnf, 0, 1, 5);
$bnf->signal_connect('clicked' => sub{ do_edit_colors($window) });


my $hbox3 = Gtk2::HBox->new(0,0);
$vbox->pack_start($hbox3, 0, 1, 5);

my $bsave=Gtk2::Button->new_with_label('Bild Speichern');
$hbox3->pack_start($bsave, 1, 1, 5);
$bsave->signal_connect('clicked' => sub{ $file=do_save($window,$file,history->get(undef,0)); });

my $bsave_all=Gtk2::Button->new_with_label('Sequenz Speichern');
$hbox3->pack_start($bsave_all, 1, 1, 5);
$bsave_all->signal_connect('clicked' => sub{
              update_history($width,$height,$bit_per_color);
              my @imgs=history->all(0);
              shift(@imgs);
              $file=do_save($window,$file,@imgs);
            });

my $b2=Gtk2::Button->new_with_label('ENDE');
$b2->signal_connect('clicked' => \&do_exit );
$vbox->pack_start($b2, 0, 1, 5);

$bnv->signal_connect('clicked' => sub{
    my ($pixbuf,$raw);
    ($pixbuf,$raw,$shift_x,$shift_y,$zoom,$max_iter)=history->next();
    $deep->value($max_iter);
    $sitem->{SHIFT_X}=$shift_x;
    $sitem->{SHIFT_Y}=$shift_y;
    $sitem->{ZOOM}=$zoom;
    do_update_hist_button($bnz,$bnv);
    $pitem->set(pixbuf=>$pixbuf);
    $pitem->raise_to_top();
    $pitem->{HAS_CHANGED}=0;
  });
$bnz->signal_connect('clicked' => sub{
    my ($pixbuf,$raw);
    ($pixbuf,$raw,$shift_x,$shift_y,$zoom,$max_iter)=history->last();
    $sitem->{SHIFT_X}=$shift_x;
    $sitem->{SHIFT_Y}=$shift_y;
    $sitem->{ZOOM}=$zoom;
    $deep->value($max_iter);
    do_update_hist_button($bnz,$bnv);
    $pitem->set(pixbuf=>$pixbuf);
    $pitem->raise_to_top();
    $pitem->{HAS_CHANGED}=0;
  });

$window->show_all;
Gtk2->main;

########################################################################
########################################################################
sub do_edit_koordinates
{
  my $win=shift;
  my @l=@{shift()};
  my $mod=0;
  my $dialog = Gtk2::Dialog->new ('KOORDIANTE', $win,
                                  [qw/modal destroy-with-parent/],
                                  'gtk-ok' => 'ok',
                                  'gtk-cancel' => 'cancel');
  my $text=Gtk2::Entry->new_with_max_length(300);
  $text->set_text(join('; ',@l));

  $dialog->get_content_area()->add($text);
  $dialog->set_default_size(300, -1);
  $dialog->set_default_response('cancel');
  $dialog->show_all();

  if('ok' eq $dialog->run)
  {
    my $txt=$text->get_text();
    $txt=~y/,/./;
    if(my @ll=split(/\s*;\s*/s,$txt))
    {
      #print "@l => @ll\n   => ";
      for(0..$#l)
      {
        if(defined($ll[$_]) && looks_like_number($ll[$_]) && $l[$_] != $ll[$_])
        {
          $mod=1;
          $l[$_]=$ll[$_];
        }
      }
    }
  }
  $dialog->destroy;
  return \@l,$mod;
}

sub do_edit_colors
{
  my $win=shift;

  my $dialog = Gtk2::Dialog->new ('FARBEN', $win, [qw/destroy-with-parent/], 'gtk-ok' => 'none');

  my $vbox=Gtk2::VBox->new(0,0);
  $dialog->get_content_area()->add($vbox);

  my $swin=Gtk2::ScrolledWindow->new();
  $vbox->pack_start($swin, 0,1,0);
  $swin->set_placement('GTK_CORNER_TOP_LEFT');
  $swin->set_policy('GTK_POLICY_ALWAYS','GTK_POLICY_NEVER');
  $swin->set_size_request(300,50);

  my $image=Gtk2::Image->new_from_pixbuf(data_to_pixbuf(colortable->example(40)));
  $swin->add_with_viewport($image);

  my $sname = colortable->signal_connect( changed => sub{
      $image->set_from_pixbuf(data_to_pixbuf(shift->example(40)));
    });

  my $br=Gtk2::Button->new_with_label('RESET');
  $br->signal_connect('clicked' => sub {
      colortable->block_signals(1);

      my $l=colortable->length()-1;
      colortable->clean();
      colortable->add($l);

      colortable->block_signals(0);
    });
  $vbox->pack_start($br, 0, 1, 5);

  my $filter = Gtk2::FileFilter->new;
  $filter->add_mime_type("text/plain");
  $filter->add_pattern("*.ctbl");


  my $bsave=Gtk2::Button->new_with_label('Speichern');
  $bsave->signal_connect('clicked' => sub {
      if(my $file=file_dialog($win,'save','FarbListe speichern','./',$filter))
      {
        $file.='.ctbl' unless($file=~/\.ctbl$/);
        colortable->save($file);
      }
    });
  $vbox->pack_start($bsave, 0, 1, 5);

  my $bload=Gtk2::Button->new_with_label('Laden');
  $bload->signal_connect('clicked' => sub {
      if(my $file=file_dialog($win,'open','FarbListe laden','./',$filter))
      {
        colortable->block_signals(1);

        my $l=colortable->length()-1;
        colortable->load($file);
        colortable->add($l);
        colortable->changed(1);

        colortable->block_signals(0);
      }
    });
  $vbox->pack_start($bload, 0, 1, 5);

  $dialog->set_default_response('ok');
  $dialog->signal_connect (response => sub{
      shift->destroy;
      colortable->signal_disconnect($sname);
    });
  $dialog->show_all;
}

########################################################################
########################################################################

sub do_save
{
  my $win=shift;
  my $file=shift;
  my @pixbufs=@_;

  my $image_filter = Gtk2::FileFilter->new;
  $image_filter->add_mime_type("image/png");
  $image_filter->add_pattern("*.png");

  my $title="Bild Speichern";
  $title="Sequenz Speichern" if(@pixbufs>1);


  if (my $f=file_dialog($win,'save',$title,$file,$image_filter))
  {
    $f=~s/((?:\.\w+?)?)$/\%03u$1/ if(@pixbufs>1 && $f!~/\%\d*[us]/);

    my $cnt=0;
    for my $pbuf (@pixbufs)
    {
      my $file=sprintf($f,$cnt++);
      _save_pixbuf($f,$pbuf);
    }
    $file=$f;
  }

  return $file;
}

sub do_update_hist_button
{
  my ($bnz,$bnv)=@_;

  if(history->pos_is_first)
  { $bnz->set_sensitive(0); }
  else
  { $bnz->set_sensitive(1); }

  if(history->pos_is_last)
  { $bnv->set_sensitive(0); }
  else
  { $bnv->set_sensitive(1); }

  return 1;
}

sub do_exit{ Gtk2->main_quit; exit; }

sub do_render
{
  my $button=shift;
  my $pitem=shift;
  my $sitem=shift;
  my $button_n=shift;
  my $button_l=shift;
  my $txt=shift;
  my $deep=shift;
  my ($width,$height,$shift_x,$shift_y,$zoom,
      $max_iter,$bit_per_color,$forks,$pack)=@_;

  $sitem->set('x1',1);
  $sitem->set('y1',1);
  $sitem->set('x2',1);
  $sitem->set('y2',1);

  $button_l->set_sensitive(0);
  $button_n->set_sensitive(0);
  $button->set_sensitive(0);
  $pitem->raise_to_top();
  my $pixbuf=$pitem->get('pixbuf');

  my $time=time();

  colortable->block_signals(1);

  if(!$pitem->{HAS_CHANGED} && int($deep->value()) == $max_iter)
  {
    if(colortable->changed)
    {
      my $hist=history->get();
      my $cnt=0;
      my $end=$width*$height;
      $txt->set_text("CREATE COLORS");
      Gtk2->main_iteration while Gtk2->events_pending;
      my $data=colortable->array_to_bin($hist->[1],sub{
          my $pos=shift;
          my $max=shift;
          $txt->set_text("UPDATE COLORS:".int(($pos/$max)*100)."%");
          Gtk2->main_iteration while Gtk2->events_pending;
        },10,$max_iter);

      $pixbuf=data_to_pixbuf($data,$width,$height,$bit_per_color);
      $hist->[0]=$pixbuf;
    }
  }
  else
  {
    $pitem->{DO_MOVE}=-1;
    my ($raw,$miter)=render_fractal($width,$height,$shift_x,$shift_y,$zoom,int($deep->value()),$txt,$forks,$pack);
    $txt->set_text("CREATE COLORS");
    Gtk2->main_iteration while Gtk2->events_pending;
    my $data=colortable->array_to_bin($raw,sub{
        my $pos=shift;
        my $max=shift;
        $txt->set_text("SET COLORS:".int(($pos/$max)*100)."%");
        Gtk2->main_iteration while Gtk2->events_pending;
      },10,$miter);

    $pixbuf=data_to_pixbuf($data,$width,$height,$bit_per_color);

    if($pitem->{HAS_CHANGED})
    {
      $sitem->{SHIFT_X}=$shift_x;
      $sitem->{SHIFT_Y}=$shift_y;
      $sitem->{ZOOM}=$zoom;
      history->add($pixbuf,$raw,$shift_x,$shift_y,$zoom,int($deep->value()));
    }
    else
    { history->update(undef,5,int($deep->value())); }
  }

  colortable->block_signals(0);

  $txt->set_text("FERTIG ($width,$height) (".(time()-$time)." Sekunden)");
  $pitem->{HAS_CHANGED}=0;
  $pitem->{DO_MOVE}=0;
  $pitem->set('pixbuf',$pixbuf);
  $button_l->set_sensitive(1) if(!history->pos_is_first);
  $button_n->set_sensitive(1) if(!history->pos_is_last);
  $button->set_sensitive(1);
  return 1;
}

sub do_select
{
  my ($item,$sitem,$event)=@_;
  $item->{DO_MOVE}=0 unless(defined($item->{DO_MOVE}));
  if ($item->{DO_MOVE}==0 && $event->type eq "button-press")
  {
    $sitem->{ZOOM}=$zoom unless(defined($sitem->{ZOOM}));
    $sitem->{SHIFT_X}=$shift_x unless(defined($sitem->{SHIFT_X}));
    $sitem->{SHIFT_Y}=$shift_y unless(defined($sitem->{SHIFT_Y}));
    $sitem->{ASPECT}=$height/$width;
    $item->{DO_MOVE}=1;
    $sitem->raise_to_top();
    $sitem->{M_X}=$event->x;
    $sitem->{M_Y}=$event->y;
    $sitem->set('x1',$event->x-1);
    $sitem->set('y1',$event->y-(1*$sitem->{ASPECT}));
    $sitem->set('x2',$event->x+1);
    $sitem->set('y2',$event->y+(1*$sitem->{ASPECT}));
    $item->{HAS_CHANGED}=1;
  }
  elsif($item->{DO_MOVE}>0 && $event->type eq "motion-notify")
  {
    my $mx=$event->x-$sitem->{M_X};
    my $my=$event->y-$sitem->{M_Y};
    $my/=$sitem->{ASPECT};

    my $mmx=0;
    my $mmy=0;
    my $mm=$mx;

    if($mx<0 && $my>=0)
    {
      $my*=-1;
      $mm=$my if($my<$mx);
      $mmy=$mm*$sitem->{ASPECT};
      $mmy*=-1;
    }
    elsif($mx>=0 && $my<0)
    {
      $my*=-1;
      $mm=$my if($my>$mx);
      $mmy=$mm*$sitem->{ASPECT};
      $mmy*=-1;
    }
    elsif($mx>=0 && $my>=0)
    {
      $mm=$my if($my>$mx);
      $mmy=$mm*$sitem->{ASPECT};
    }
    elsif($mx<0 && $my<0)
    {
      $mm=$my if($my<$mx);
      $mmy=$mm*$sitem->{ASPECT};
    }

    $mmx=$mm;

    $sitem->set('x2',$sitem->{M_X}+$mmx);
    $sitem->set('y2',$sitem->{M_Y}+$mmy);
    $sitem->set('x1',$sitem->{M_X}-$mmx);
    $sitem->set('y1',$sitem->{M_Y}-$mmy);
  }
  elsif($item->{DO_MOVE}>0 && $event->type eq "button-release")
  {
    $item->{DO_MOVE}=0;

    my $x1=$sitem->get('x1');
    my $y1=$sitem->get('y1');
    my $x2=$sitem->get('x2');
    my $y2=$sitem->get('y2');
    ($x1,$x2)=($x2,$x1) if($x1>$x2);
    ($y1,$y2)=($y2,$y1) if($y1>$y2);

    return if($x2-$x1<=0 || $y2-$y1<=0);

    $shift_x=$sitem->{SHIFT_X}-$x1/$sitem->{ZOOM};
    $shift_y=$sitem->{SHIFT_Y}-$y1/$sitem->{ZOOM};
    my $z1=($width*$sitem->{ZOOM})/abs($x2-$x1);
    my $z2=($height*$sitem->{ZOOM})/abs($y2-$y1);
    $zoom=($z1+$z2)/2;
  }
  return 1;
}

#-----------------------------------------------------------------------

sub file_dialog
{
  my $win=shift;
  my $type=shift // 'save';
  my $title=shift // 'Speichern';
  my $file=shift;
  my $filter=shift;


  my $bimg='gtk-save';
  $bimg='gtk-open' if($type eq 'open');


  my $dialog=Gtk2::FileChooserDialog->new ($title, $win, $type,
                                           'gtk-cancel' => 'cancel',
                                           $bimg => 'ok');
  $dialog->set_default_response ('cancel');
  $dialog->set_filter($filter) if($filter);
  if($file)
  {
    if(-f $file)
    {
      my ($volume,$directories,$file) = File::Spec->splitpath( $file );
      my $dir=File::Spec->join($volume,$directories);
      $dialog->set_current_folder($dir);
      $dialog->set_current_name($file);
    }
    elsif(-d $file)
    { $dialog->set_current_folder($file); }
    else
    {
      my ($volume,$directories,$file) = File::Spec->splitpath( $file );
      my $dir=File::Spec->join($volume,$directories);
      $dialog->set_current_folder($dir) if(-d $dir);
    }
  }

  if ('ok' eq $dialog->run ())
  { $file=$dialog->get_filename(); }
  else
  { $file=undef; }

  if($file && $type eq 'save' && -f $file)
  {
    my $dialog = Gtk2::MessageDialog->new ($win, 'destroy-with-parent', 'question',
                                      'yes-no',  "Datei %s überscreiben?", $file);
    $file=undef if ('yes' ne $dialog->run);
    $dialog->destroy;
  }

  $dialog->destroy;
  return $file;
}

#-----------------------------------------------------------------------

sub _save_pixbuf
{
  my $filename=shift;
  my $pixbuf=shift;
  $pixbuf->save($filename, 'png');
}

sub data_to_pixbuf
{
 my ($data,$b,$h,$bpc)=@_;
 return Gtk2::Gdk::Pixbuf->new_from_data($data,'GDK_COLORSPACE_RGB',0,$bpc,$b,$h,$b*3);
}

#-----------------------------------------------------------------------

sub update_history
{
  my ($width,$height,$bit_per_color)=@_;
  if(colortable->changed)
  {
    for my $p (1..history->length()-1)
    {
      my $data=colortable->array_to_bin(history->get($p,1));
      history->update($p,1,data_to_pixbuf($data,$width,$height,$bit_per_color));
    }
    colortable->changed(0);
  }
}

#-----------------------------------------------------------------------

sub render_fractal
{
  my $mx=shift;
  my $my=shift;
  my $sx=shift;
  my $sy=shift;
  my $zoom=shift;
  my $mi=shift;
  my $txt=shift;
  my $forks=shift || 1;
  my $deep=0;

  $txt->set_text("!RENDERE!");
  my $data='';

  # nicht forken
  if($forks<=1)
  {
    for my $y (1..$my)
    {
      $txt->set_text("BEARBEITE ZEILE: $y von $my");
      Gtk2->main_iteration while Gtk2->events_pending;
      for my $x (1..$mx)
      {
        my $i=iterate_mandelbrot($x, $sx, $y, $sy, $zoom, $mi);
        $deep=$i if($deep<$i);
        $data.=pack($raw_pack, $i);
      }
    }
  }
  # forken
  else
  {
    my @data_list;

    my @chunks;
    my $size=$my/$forks;
    push(@chunks,[int($size*($_-1)),int($size*$_)]) for(1..$forks);
    push(@chunks,[int($size*$forks),$my]) if(int($size*$forks)<$my);

    while(@chunks)
    {
      my $d=shift(@chunks);
      $txt->set_text("Starte Fork für Zeilen $$d[0] - $$d[1]");
      Gtk2->main_iteration while Gtk2->events_pending;
      my $pid=open(my $pipe, '-|');
      die("error fork\n") unless( defined ($pid) );
      if($pid)
      { push(@data_list,[$pid,$pipe,'']); }
      else
      {
        #child process
        my $deep=0;
        my $yy='';
        for my $y($d->[0]..$d->[1])
        {
          for my $x (1..$mx)
          {
            my $i=iterate_mandelbrot($x, $sx, $y, $sy, $zoom, $mi);
            $deep=$i if($deep<$i);
            $yy.=pack($raw_pack, $i);
          }
        }
        print "$yy\n";
        print "---ENDE---\n$deep";
        exit();
      }
    }

    $txt->set_text("Warte auf Forks");
    Gtk2->main_iteration while Gtk2->events_pending;

    my $run=1;
    while($run)
    {
      $run=0;
      for my $chunk (@data_list)
      {
        my $pid=$chunk->[0];
        if($pid)
        {
          $run=1;
          eval
          {
            $SIG{ALRM}=sub{ die(); };
            my $fh=$chunk->[1];
            my $line='';
            $txt->set_text("FORK: $chunk->[0] lese Zeilen");
            Gtk2->main_iteration while Gtk2->events_pending;
            while(defined($line))
            {
              alarm(1);
              $line=<$fh>;
              alarm(0);
              if(!defined($line) || $line eq "---ENDE---\n")
              {
                $chunk->[1]=undef;
                $chunk->[0]=undef;
                if(defined($line))
                {
                  my $p=<$fh>;
                  chomp($p);
                  $deep=$p if($deep<$p);
                }
                close($fh);
                last;
              }
              $chunk->[2].=$line;
            }
            $txt->set_text("FORK: $pid $. Zeilen gelesen");
            Gtk2->main_iteration while Gtk2->events_pending;
            alarm(0);
          };
        }
      }
    }

    for (@data_list)
    {
      my $d=$_->[2];
      chomp($d);
      $data.=$d;
    }
  }

  return $data,$deep;
}

sub _iterate_mandelbrot_perl
{
  my ($x, $sx, $y, $sy, $zoom, $mi)=@_;
  $x/=$zoom;
  $y/=$zoom;
  $x-=$sx;
  $y-=$sy;
  my $i=0;
  my $betr=0;
  my $rx=0;
  my $ry=0;
  my $rxn=0;
  my $ryn=0;

  do
  {
    $rxn = $rx**2 - $ry**2 + $x;
    $ryn = (2*$rx*$ry) + $y;
    $betr= $rxn**2 + $ryn**2;
    $rx  = $rxn;
    $ry  = $ryn;
    return $i if($betr > 4);
    $i++;
  }
  while($i < $max_iter);

  $i=$i-log(log($betr)*log(4))*log(2) if($colortable::use_percent);

  return $i;
}

__DATA__

__C__

# include <math.h>

double _iterate_mandelbrot_double_c(double inx, double sx, double iny, double sy, double zoom, long max_iter)
{
  double i=0;

  long double x=(inx/zoom)-sx;
  long double y=(iny/zoom)-sy;
  long double betr=0;
  long double rx=0;
  long double ry=0;
  long double rxn=0;
  long double ryn=0;

  do
  {
    rxn = (rx*rx) - (ry*ry) + x;
    ryn = (2*rx*ry) + y;
    betr= (rxn*rxn) + (ryn*ryn);
    rx  = rxn;
    ry  = ryn;
    if(betr > 4) return i;
    i++;
  }
  while(i < max_iter);

  i=i-log(log(betr)*log(4))*log(2);

  return i;
}

long _iterate_mandelbrot_long_c(double inx, double sx, double iny, double sy, double zoom, long max_iter)
{
  long i=0;

  long double x=(inx/zoom)-sx;
  long double y=(iny/zoom)-sy;
  long double betr=0;
  long double rx=0;
  long double ry=0;
  long double rxn=0;
  long double ryn=0;

  do
  {
    rxn = (rx*rx) - (ry*ry) + x;
    ryn = (2*rx*ry) + y;
    betr= (rxn*rxn) + (ryn*ryn);
    rx  = rxn;
    ry  = ryn;
    if(betr > 4) return i;
    i++;
  }
  while(i < max_iter);

  return i;
}

